#!/usr/bin/env python3

import argparse
import subprocess
import os
import json
import math

from jinja2 import Environment, FileSystemLoader, select_autoescape


DOCKER_COMPOSE_TEMPLATE = "docker-compose-local.yml.j2"
DOCKER_COMPOSE_OUTPUT_FILE = "docker-compose-local.generated.yml"

def spawn_local_game(game_config_path, secrets_path):
    if not os.path.exists(game_config_path):
        print("The specified game_config path does not exist")
        return

    if not os.path.exists(secrets_path):
        print("The specified path to the secret folder does not exist")
        return

    with open(game_config_path, 'r') as f:
        game_config = json.load(f)

    with open(os.path.join(secrets_path, 'database-api', 'secret'), 'r') as f:
        api_secret = f.read().strip()

    num_scriptbots = int(math.ceil(len(game_config["teams"]) / 10.0))
    num_teams = len(game_config['teams']) 
    active_services = [service['name'] for service in game_config['services'] if service['state'] == 'enabled']

    jinja_env = Environment(
        loader=FileSystemLoader(searchpath='./'),
        autoescape=select_autoescape(['yml'])
    )

    teamvm_entrypoints = []

    for team_id in range(1, len(game_config['teams'] ) + 1):
        teamvm_entry = []
        for service_id, service_name in enumerate(active_services):
            exposed_port = 10000 + service_id + 1
            portforward_cmd = "socat TCP-LISTEN:{},fork TCP:chall_{}_{}:6666".format(exposed_port, service_name, team_id)
            teamvm_entry.append(portforward_cmd)
        
        final_teamvm_entry = " & ".join(teamvm_entry)
        teamvm_entrypoints.append(final_teamvm_entry)

    template = jinja_env.get_template(DOCKER_COMPOSE_TEMPLATE)

    with open(DOCKER_COMPOSE_OUTPUT_FILE, 'w') as out_f:
        out_f.write(template.render(
            num_scriptbots=num_scriptbots,
            num_teams=num_teams,
            api_secret=api_secret,
            # services_ids=range(1, len(active_services) + 1)
            service_names=active_services,
            teamvm_entrypoints=teamvm_entrypoints
        ))
    
    print('''
Configuration for docker-compose successfully generated in ./{0}

Spawn the infrastructure locally with the following command:
    - docker-compose -f {0} up

Destroy the infrastructure locally with the following command:
    - docker-compose -f {0} down -v --remove-orphans

To start the game: http://localhost:5000/game/insert?secret={1}
    '''.format(DOCKER_COMPOSE_OUTPUT_FILE, api_secret))


if __name__ == '__main__':
    argparser = argparse.ArgumentParser()
    argparser.add_argument('game_config', type=str, help="Path to game_config.json file")
    argparser.add_argument('secrets_path', type=str, help="Path to the folder generated by make_secret.sh")
    args = argparser.parse_args()

    spawn_local_game(args.game_config, args.secrets_path)
